! ********************************************** 
!   04/01/2008 
! ********************************************** 
! 
! Main program for structured eigenvalue computation
! (QR with single shift). This is companion software
! for the paper "A fast implicit QR eigenvalue 
! algorithm for companion matrices" by D.A.Bini,
! P.Boito, Y.Eidelman, L.Gemignani and I.Gohberg
! (preprint n. 7.26.1700, November 2008, Department
! of Mathematics, University of Pisa).
!
! Includes: 
!
! - comparison with the computation time of Lapack,
! - comparison with the eigenvalues found by Lapack
!   (using both the "minimum distance" and the "sorting"
!   criteria).
!
! Please note: 
! - the subroutine sort_eig has been taken
!   from the program by Xia,
! - the subroutine ssort.f is needed when compiling
!   this program (easy to find on the net).  
! 
 
 
program main 
   
  implicit none 
   
  complex(kind(1d0)), allocatable :: eig(:),myeig(:),A(:,:),poly(:)
  complex(kind(1d0)), allocatable :: work(:),vl(:,:),vr(:,:) 
  real(kind(1d0)) :: w1,w2,w3,our_time,maxdiffeR,maxdiffeI,coef(20)
  real(kind(1d0)), allocatable :: rwork(:),dist1(:),dist2(:) 
  real(kind(1d0)), allocatable :: eR1(:),eR2(:),eI1(:),eI2(:),seR1(:),seR2(:),seI1(:),seI2(:),diffeR(:),diffeI(:)
   
  integer :: i,j,n,lwork,info,lapacksize 
  integer :: clock_start,clock_end,clock_rate 
   
  write(*,"(a)",advance="no") "Degree = " 
  read(*,*) n 
  write(*,"(a)",advance="no") "Lapacksize = " 
  read(*,*) lapacksize 
   
  allocate(poly(n),myeig(n)) 
   
  ! Form monic polynomial with random coefficients 
  ! (poly contains only the nonleading coefficients) 
   
  call random_seed
   
  do i=1,n 
     call random_number(w1) 
     call random_number(w2) 
    poly(i)=(2.d0,0.d0)*w1-(1.d0, 0.d0)+(0.d0,1.d0)*w2-(0.d0, 1.d0) 
  enddo 

! Roots of 1:
!!$poly(1:n-1)=0.d0
!!$poly(n)=(-1.d0,0.d0)

!**** Uncomment the following to save the polynomial
!  open(unit=300,file="poly.txt") 
!  do j=1,n 
!write(300,*) poly(j) 
!enddo 
!close(300) 

  ! Apply begg_eig.f90 to POLY and compute running time


  CALL SYSTEM_CLOCK(COUNT_RATE=clock_rate) 
  CALL SYSTEM_CLOCK(COUNT=clock_start) 

  call begg_eig_lapack(n,poly,lapacksize,myeig)


  CALL SYSTEM_CLOCK(COUNT=clock_end)  
  our_time=(REAL(clock_end - clock_start)/clock_rate) 
  write(*,*)"Our time = ",our_time 

  ! Apply the Lapack routine ZGEEV to the companion matrix A
  ! associated with POLY and compute running time
   
  allocate(vL(1,n),vR(1,n),A(n,n),eig(n),dist1(n),dist2(n)) 
   
  A=0.d0 
  do i=1,n-1 
     A(i+1,i)=(1.d0,0.d0) 
  enddo 
  A(1,1:n)=-poly 
   
  LWORK = max(1,3*n+100) 
  allocate (WORK(LWORK)) 
  allocate(rwork(2*n)) 
  CALL SYSTEM_CLOCK(COUNT=clock_start)
  call ZGEEV('n','n',n,A,n,eig,vL,1,vR,1,WORK,LWORK,RWORK,INFO) 

  CALL SYSTEM_CLOCK(COUNT=clock_end)  
  our_time=(REAL(clock_end - clock_start)/clock_rate) 
  write(*,*)"Lapack time = ",our_time


  ! Compare eigenvalues:
  !
  ! comparison by minimum distance  

  compare_outer:do j=1,n 
     compare_inner:do i=1,n 
        dist2(i)=abs(myeig(j)-eig(i)) 
     end do compare_inner 
     dist1(j)=minval(dist2) 
  end do compare_outer 
  write(*,*)"distance = ",maxval(dist1) 
   

  ! comparison by sorting
  allocate(eR1(n),eR2(n),eI1(n),eI2(n),seR1(n),seR2(n),seI1(n),seI2(n),diffeR(n),diffeI(n))

  eR1=real(myeig)
  eI1=aimag(myeig)
  eR2=real(eig)
  eI2=aimag(eig)

  call sort_eig(n,eR1,eI1,seR1,seI1)
  call sort_eig(n,eR2,eI2,seR2,seI2)

  diffeR = abs(seR2 - seR1)
  diffeI = abs(seI2 - seI1)
  maxdiffeR = maxval(abs(diffeR))
  maxdiffeI = maxval(abs(diffeI))

  write(*,*)"distance (by sorting) = ",maxdiffeR, maxdiffeI
  !write(*,*)"abs. distance (by sorting) = ",sqrt(maxdiffeR**2 + maxdiffeI**2)

end program main 
!contains
 

subroutine	sort_eig(n,eR,eI,seR,seI)
  !!
  !!	sort the sequence according to real part first, if two complex numbers
  !!		have the same real part, put the one with bigger imaginary part in
  !!		front
  !!

  implicit none
  integer				n
  double precision	eR(n),eI(n),seR(n),seI(n),tmpR,tmpI
  integer				i,IY(n),KFLAG,tmpindx

  do i = 1,n
     IY(i) = i
  enddo

  seR = eR
  call ssort(seR,iy,n,kflag)

  do i = 1,n
     seI(i) = eI(iy(i))
  enddo

  tmpR = seR(1)
  tmpI = seI(1)
  do i = 2,n
     if (tmpR == seR(i)) then
        if (tmpI < seI(i)) then
           seI(i-1) = seI(i)
           seI(i)	= tmpI
           tmpindx	= IY(i-1)
           IY(i-1)	= IY(i)
           IY(i) = tmpindx
        endif
     else
        tmpR = seR(i)
        tmpI = seI(i)
     endif
  enddo

  return

end subroutine sort_eig

